@@ -10,13 +10,13 @@ import (
 	"os"
 	"regexp"
 	"strings"
+	"sync"
 	"testing"
 	"time"
 
 	"github.com/aws/aws-sdk-go/aws"
 	"github.com/aws/aws-sdk-go/aws/defaults"
 	"github.com/aws/aws-sdk-go/aws/endpoints"
-	E "github.com/olivere/elastic/v7"
 	"github.com/stretchr/testify/require"
 	"gitlab.com/gitlab-org/gitlab-elasticsearch-indexer/internal/mode/advanced/elastic"
 )
@@ -145,7 +145,6 @@ func TestResolveAWSCredentialsEc2RoleProfiles(t *testing.T) {
 		))
 		require.NoError(t, err)
 
-		// Bypass shared aws credential config file
 		os.Setenv("AWS_SHARED_CREDENTIALS_FILE", "/tmp/notexist")
 		defer os.Unsetenv("AWS_SHARED_CREDENTIALS_FILE")
 
@@ -175,7 +174,6 @@ func TestResolveAWSCredentialsECSCredsProvider(t *testing.T) {
 	))
 	require.NoError(t, err)
 
-	// Bypass shared aws credential config file
 	os.Setenv("AWS_SHARED_CREDENTIALS_FILE", "/tmp/notexist")
 	os.Setenv("AWS_CONTAINER_CREDENTIALS_FULL_URI", server.URL+"/latest/meta-data/iam/security-credentials/RoleName")
 	defer func() {
@@ -261,87 +259,16 @@ func setupTestClient(t *testing.T) *elastic.Client {
 
 func setupTestClientAndCreateIndex(t *testing.T) *elastic.Client {
 	client := setupTestClient(t)
+
+	_ = client.DeleteIndex(client.IndexNameDefault)
+	_ = client.DeleteIndex(client.IndexNameCommits)
+
 	require.NoError(t, client.CreateDefaultWorkingIndex())
 	require.NoError(t, client.CreateCommitsWorkingIndex())
 
 	return client
 }
 
-func putTestAlias(t *testing.T, c *elastic.Client, params *elastic.CreateAliasParams) {
-	require.NoError(t, c.CreateAlias(params))
-}
-
-type testDoc struct {
-	ProjectId string `json:"project_id"`
-}
-
-func TestDeleteFromRolledOverIndices(t *testing.T) {
-	client := setupTestClient(t)
-	client.SearchCuration = true
-
-	writeIndexName := client.IndexNameDefault
-	rolloverIndexName := "test-index-1"
-
-	aliasName := "test-alias"
-	docId := "1"
-	doc := &testDoc{ProjectId: fmt.Sprintf("%v", client.ParentID())}
-	r := fmt.Sprintf("project_%v", doc.ProjectId)
-
-	// index a doc in rollover index
-	if _, err := client.Client.Index().Index(rolloverIndexName).Routing(r).Id(docId).BodyJson(doc).Do(context.TODO()); err != nil {
-		t.Fatal(err)
-	}
-
-	// duplicate same doc in write index
-	if _, err := client.Client.Index().Index(writeIndexName).Routing(r).Id(docId).BodyJson(doc).Do(context.TODO()); err != nil {
-		t.Fatal(err)
-	}
-
-	// create write + read alias
-	putTestAlias(t, client, &elastic.CreateAliasParams{Index: writeIndexName, Alias: aliasName, IsWriteIndex: true})
-	putTestAlias(t, client, &elastic.CreateAliasParams{Index: rolloverIndexName, Alias: aliasName, IsWriteIndex: false})
-
-	// cleanup after test
-	defer func() {
-		require.NoError(t, client.RemoveAlias(&elastic.RemoveAliasParams{
-			Index: writeIndexName, Alias: aliasName,
-		}))
-		require.NoError(t, client.RemoveAlias(&elastic.RemoveAliasParams{
-			Index: rolloverIndexName, Alias: aliasName,
-		}))
-		require.NoError(t, client.DeleteIndex(writeIndexName))
-		require.NoError(t, client.DeleteIndex(rolloverIndexName))
-	}()
-
-	require.NoError(t, client.DeleteFromRolledOverIndices(&elastic.DeleteFromRolledOverIndicesParams{
-		AliasName: aliasName,
-		DocType:   "project",
-		DocId:     docId,
-	}))
-	require.NoError(t, client.Flush())
-
-	// refresh both indices
-	if _, err := client.Client.Refresh().Index(rolloverIndexName, writeIndexName).Do(context.TODO()); err != nil {
-		t.Fatal(err)
-	}
-
-	search, searchErr := client.Client.Search().Index(aliasName).Query(E.NewMatchAllQuery()).Do(context.TODO())
-	require.NoError(t, searchErr)
-
-	if search.TotalHits() != 1 {
-		t.Errorf("expected hits == 1; got %v", search.TotalHits())
-	}
-
-	var d testDoc
-	hit := search.Hits.Hits[0]
-
-	if err := json.Unmarshal(hit.Source, &d); err != nil {
-		t.Fatal(err) // serialization error
-	}
-
-	require.Equal(t, writeIndexName, hit.Index, "expected document to only be in write index")
-}
-
 func TestElasticClientIndexAndRetrieval(t *testing.T) {
 	client := setupTestClientAndCreateIndex(t)
 
@@ -512,3 +439,364 @@ func TestHealthcheckIsDisabled(t *testing.T) {
 
 	require.Nil(t, req)
 }
+
+func TestBulkSizeTracking(t *testing.T) {
+	originalConfig := os.Getenv("ELASTIC_CONNECTION_INFO")
+	defer os.Setenv("ELASTIC_CONNECTION_INFO", originalConfig)
+
+	var configMap map[string]interface{}
+	if originalConfig != "" {
+		err := json.Unmarshal([]byte(originalConfig), &configMap)
+		require.NoError(t, err)
+	} else {
+		configMap = make(map[string]interface{})
+	}
+
+	configMap["max_bulk_size_bytes"] = 1000
+	modifiedConfig, err := json.Marshal(configMap)
+	require.NoError(t, err)
+
+	os.Setenv("ELASTIC_CONNECTION_INFO", string(modifiedConfig))
+
+	client := setupTestClient(t)
+	require.NotNil(t, client)
+	defer client.Close()
+
+	require.Equal(t, 0, client.CurrentBatchSize())
+
+	smallDoc := map[string]interface{}{"field": "value", "field2": "value2"}
+
+	initialSize := client.CurrentBatchSize()
+	client.Index("blob", projectIDString+"_0000", smallDoc)
+	require.Greater(t, client.CurrentBatchSize(), initialSize)
+
+	err = client.Flush()
+	require.NoError(t, err)
+	require.Equal(t, 0, client.CurrentBatchSize())
+}
+
+func TestProactiveFlushOnSizeLimit(t *testing.T) {
+	originalConfig := os.Getenv("ELASTIC_CONNECTION_INFO")
+	defer os.Setenv("ELASTIC_CONNECTION_INFO", originalConfig)
+
+	var configMap map[string]interface{}
+	if originalConfig != "" {
+		err := json.Unmarshal([]byte(originalConfig), &configMap)
+		require.NoError(t, err)
+	} else {
+		configMap = make(map[string]interface{})
+	}
+
+	configMap["max_bulk_size_bytes"] = 500
+
+	modifiedConfig, err := json.Marshal(configMap)
+	require.NoError(t, err)
+	os.Setenv("ELASTIC_CONNECTION_INFO", string(modifiedConfig))
+
+	client := setupTestClient(t)
+	require.NotNil(t, client)
+	defer client.Close()
+
+	largeDoc := map[string]interface{}{
+		"content": strings.Repeat("a", 200),
+	}
+
+	client.Index("blob", projectIDString+"_test_1", largeDoc)
+	size1 := client.CurrentBatchSize()
+	require.NotEmpty(t, size1)
+	require.Less(t, size1, 500, "First document should fit within limit")
+
+	client.Index("blob", projectIDString+"_test_2", largeDoc)
+	size2 := client.CurrentBatchSize()
+
+	require.NotEmpty(t, size2)
+	require.Less(t, size2, 500, "Batch size should be less than max_bulk_size_bytes")
+
+	err = client.Flush()
+	require.NoError(t, err)
+	require.Equal(t, 0, client.CurrentBatchSize())
+}
+
+func TestRemoveBulkSizeTracking(t *testing.T) {
+	originalConfig := os.Getenv("ELASTIC_CONNECTION_INFO")
+	defer os.Setenv("ELASTIC_CONNECTION_INFO", originalConfig)
+
+	var configMap map[string]interface{}
+	if originalConfig != "" {
+		err := json.Unmarshal([]byte(originalConfig), &configMap)
+		require.NoError(t, err)
+	} else {
+		configMap = make(map[string]interface{})
+	}
+
+	configMap["max_bulk_size_bytes"] = 1000
+	modifiedConfig, err := json.Marshal(configMap)
+	require.NoError(t, err)
+
+	os.Setenv("ELASTIC_CONNECTION_INFO", string(modifiedConfig))
+
+	client := setupTestClient(t)
+	require.NotNil(t, client)
+	defer client.Close()
+
+	require.Equal(t, 0, client.CurrentBatchSize())
+
+	client.Remove("blob", "test_id_1")
+
+	require.NotEmpty(t, client.CurrentBatchSize())
+
+	initialSize := client.CurrentBatchSize()
+
+	client.Remove("blob", "test_id_2")
+
+	require.Greater(t, client.CurrentBatchSize(), initialSize)
+
+	err = client.Flush()
+	require.NoError(t, err)
+	require.Equal(t, 0, client.CurrentBatchSize())
+}
+
+func TestDeleteBulkSizeTracking(t *testing.T) {
+	originalConfig := os.Getenv("ELASTIC_CONNECTION_INFO")
+	defer os.Setenv("ELASTIC_CONNECTION_INFO", originalConfig)
+
+	var configMap map[string]interface{}
+	if originalConfig != "" {
+		err := json.Unmarshal([]byte(originalConfig), &configMap)
+		require.NoError(t, err)
+	} else {
+		configMap = make(map[string]interface{})
+	}
+
+	configMap["max_bulk_size_bytes"] = 1000
+	modifiedConfig, err := json.Marshal(configMap)
+	require.NoError(t, err)
+
+	os.Setenv("ELASTIC_CONNECTION_INFO", string(modifiedConfig))
+
+	client := setupTestClient(t)
+	require.NotNil(t, client)
+	defer client.Close()
+
+	largeDoc := map[string]interface{}{
+		"content": strings.Repeat("a", 200),
+	}
+	client.Index("blob", projectIDString+"_test_1", largeDoc)
+	client.Index("blob", projectIDString+"_test_2", largeDoc)
+	err = client.Flush()
+	require.NoError(t, err)
+
+	deleteParams := &elastic.DeleteParams{
+		Index:   configMap["index_name"].(string),
+		Routing: projectIDString,
+		DocId:   projectIDString + "_test_1",
+	}
+
+	client.Delete(deleteParams)
+	require.NotEmpty(t, client.CurrentBatchSize())
+
+	initialSize := client.CurrentBatchSize()
+
+	deleteParams2 := &elastic.DeleteParams{
+		Index:   configMap["index_name"].(string),
+		Routing: projectIDString,
+		DocId:   projectIDString + "_test_2",
+	}
+	client.Delete(deleteParams2)
+	require.Greater(t, client.CurrentBatchSize(), initialSize)
+
+	err = client.Flush()
+	require.NoError(t, err)
+	require.Equal(t, 0, client.CurrentBatchSize())
+}
+
+func TestMixedOperationsBulkSizeTracking(t *testing.T) {
+	originalConfig := os.Getenv("ELASTIC_CONNECTION_INFO")
+	defer os.Setenv("ELASTIC_CONNECTION_INFO", originalConfig)
+
+	var configMap map[string]interface{}
+	if originalConfig != "" {
+		err := json.Unmarshal([]byte(originalConfig), &configMap)
+		require.NoError(t, err)
+	} else {
+		configMap = make(map[string]interface{})
+	}
+
+	configMap["max_bulk_size_bytes"] = 1000
+	modifiedConfig, err := json.Marshal(configMap)
+	require.NoError(t, err)
+
+	os.Setenv("ELASTIC_CONNECTION_INFO", string(modifiedConfig))
+
+	client := setupTestClient(t)
+	require.NotNil(t, client)
+	defer client.Close()
+
+	doc := map[string]interface{}{"field": "value"}
+	client.Index("blob", "test_id_1", doc)
+	size1 := client.CurrentBatchSize()
+
+	client.Remove("blob", "test_id_2")
+	size2 := client.CurrentBatchSize()
+	require.Greater(t, size2, size1)
+
+	deleteParams := &elastic.DeleteParams{
+		Index:   configMap["index_name"].(string),
+		Routing: projectIDString,
+		DocId:   projectIDString + "_test_2",
+	}
+	client.Delete(deleteParams)
+	size3 := client.CurrentBatchSize()
+	require.Greater(t, size3, size2)
+
+	err = client.Flush()
+	require.NoError(t, err)
+	require.Equal(t, 0, client.CurrentBatchSize())
+}
+
+func TestConcurrentOperationsThreadSafety(t *testing.T) {
+	originalConfig := os.Getenv("ELASTIC_CONNECTION_INFO")
+	defer os.Setenv("ELASTIC_CONNECTION_INFO", originalConfig)
+
+	var configMap map[string]interface{}
+	if originalConfig != "" {
+		err := json.Unmarshal([]byte(originalConfig), &configMap)
+		require.NoError(t, err)
+	} else {
+		configMap = make(map[string]interface{})
+	}
+
+	configMap["max_bulk_size_bytes"] = 5000
+
+	modifiedConfig, err := json.Marshal(configMap)
+	require.NoError(t, err)
+	os.Setenv("ELASTIC_CONNECTION_INFO", string(modifiedConfig))
+
+	client := setupTestClient(t)
+	require.NotNil(t, client)
+	defer client.Close()
+
+	const numGoroutines = 10
+	const operationsPerGoroutine = 20
+
+	var wg sync.WaitGroup
+	wg.Add(numGoroutines)
+
+	errorChan := make(chan error, numGoroutines)
+
+	for i := 0; i < numGoroutines; i++ {
+		go func(goroutineID int) {
+			defer wg.Done()
+			defer func() {
+				if r := recover(); r != nil {
+					errorChan <- fmt.Errorf("goroutine %d panicked: %v", goroutineID, r)
+				}
+			}()
+
+			for j := 0; j < operationsPerGoroutine; j++ {
+				doc := map[string]interface{}{
+					"content":      fmt.Sprintf("content from goroutine %d operation %d", goroutineID, j),
+					"goroutine_id": goroutineID,
+					"operation_id": j,
+				}
+
+				switch j % 4 {
+				case 0:
+					client.Index("blob", fmt.Sprintf("%s_concurrent_%d_%d", projectIDString, goroutineID, j), doc)
+				case 1:
+					client.Remove("blob", fmt.Sprintf("%s_remove_%d_%d", projectIDString, goroutineID, j))
+				case 2:
+					client.Delete(&elastic.DeleteParams{
+						Index:   client.IndexNameDefault,
+						Routing: fmt.Sprintf("project_%v", client.ProjectID),
+						DocId:   fmt.Sprintf("delete_%d_%d", goroutineID, j),
+					})
+				case 3:
+					if j%10 == 0 {
+						client.Flush()
+					} else {
+						client.Index("blob", fmt.Sprintf("%s_index_%d_%d", projectIDString, goroutineID, j), doc)
+					}
+				}
+
+				if j%5 == 0 {
+					time.Sleep(1 * time.Millisecond)
+				}
+			}
+		}(i)
+	}
+
+	wg.Wait()
+	close(errorChan)
+
+	for err := range errorChan {
+		t.Errorf("Concurrent operation failed: %v", err)
+	}
+
+	require.NoError(t, client.Flush())
+	require.Equal(t, 0, client.CurrentBatchSize())
+}
+
+func TestConcurrentFlushOperations(t *testing.T) {
+	originalConfig := os.Getenv("ELASTIC_CONNECTION_INFO")
+	defer os.Setenv("ELASTIC_CONNECTION_INFO", originalConfig)
+
+	var configMap map[string]interface{}
+	if originalConfig != "" {
+		err := json.Unmarshal([]byte(originalConfig), &configMap)
+		require.NoError(t, err)
+	} else {
+		configMap = make(map[string]interface{})
+	}
+
+	configMap["max_bulk_size_bytes"] = 2000
+
+	modifiedConfig, err := json.Marshal(configMap)
+	require.NoError(t, err)
+	os.Setenv("ELASTIC_CONNECTION_INFO", string(modifiedConfig))
+
+	client := setupTestClient(t)
+	require.NotNil(t, client)
+	defer client.Close()
+
+	doc := map[string]interface{}{"content": "test content"}
+	for i := 0; i < 5; i++ {
+		client.Index("blob", fmt.Sprintf("%s_flush_test_%d", projectIDString, i), doc)
+	}
+
+	const numFlushGoroutines = 5
+	var wg sync.WaitGroup
+	wg.Add(numFlushGoroutines)
+
+	errorChan := make(chan error, numFlushGoroutines)
+
+	for i := 0; i < numFlushGoroutines; i++ {
+		go func(goroutineID int) {
+			defer wg.Done()
+			defer func() {
+				if r := recover(); r != nil {
+					errorChan <- fmt.Errorf("flush goroutine %d panicked: %v", goroutineID, r)
+				}
+			}()
+
+			for j := 0; j < 3; j++ {
+				err := client.Flush()
+				if err != nil {
+					t.Logf("Flush error in goroutine %d: %v", goroutineID, err)
+				}
+				time.Sleep(10 * time.Millisecond)
+			}
+		}(i)
+	}
+
+	wg.Wait()
+	close(errorChan)
+
+	for err := range errorChan {
+		if strings.Contains(err.Error(), "panicked") {
+			t.Errorf("Concurrent flush operation failed: %v", err)
+		}
+	}
+
+	require.Equal(t, 0, client.CurrentBatchSize())
+}
