@@ -16,6 +16,16 @@ const (
 // Set the default chunk size to 1000
 const DefaultChunkSize = 1000
 
+// ChunkStrategy represents the algorithm used for chunking files
+type ChunkStrategy string
+
+const (
+	// ChunkStrategyCodeBytes parses code to select chunk split points and uses byte based chunk size limits. `chunk_size` sets maximum number of bytes in each chunk.
+	ChunkStrategyCodeBytes ChunkStrategy = "code_bytes"
+	// ChunkStrategyCodePreBert parses code to select chunk split points and uses BERT pre-tokenization based chunk size limits. `chunk_size` sets the maximum number of BERT tokens in each chunk.
+	ChunkStrategyCodePreBert ChunkStrategy = "code_pre_bert"
+)
+
 // ChunkOptions contains configuration for the chunking process
 type ChunkOptions struct {
 	ChunkSize    uint16 `json:"chunk_size"`
